from __future__ import annotations

import logging
import socket
import subprocess
import threading
import time
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from io import FileIO
from queue import Queue
from typing import TYPE_CHECKING, Any, TypedDict, cast

import urllib3
from tenacity import retry, retry_if_exception, retry_if_exception_type, stop_after_attempt, wait_fixed
from typing_extensions import NotRequired

from dagster_ray.kuberay.client.base import BaseKubeRayClient, load_kubeconfig
from dagster_ray.kuberay.utils import is_retryable_k8s_api_exception, k8s_service_fqdn

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from kubernetes.client import ApiClient
    from ray.job_submission import JobSubmissionClient


GROUP = "ray.io"
VERSION = "v1"
PLURAL = "rayclusters"
KIND = "RayCluster"


def enqueue_output(out: FileIO, queue: Queue, should_stop: Callable[[], bool]):
    import select

    try:
        while True:
            if should_stop():
                break
            rlist, _, _ = select.select([out], [], [], 1.0)
            if not rlist:
                continue
            line = out.readline()
            if not line:
                break
            queue.put(line)
    except Exception:
        logger.exception("Exception encountered while reading port-forwarded RayCluster logs")
    finally:
        try:
            out.close()
        except Exception:
            logger.exception("Exception encountered while closing RayCluster logs during port-forwarding")


def get_random_available_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))  # Bind to an available port chosen by the OS
        _, port = s.getsockname()  # Get the port number assigned
        return port


class RayClusterEndpoints(TypedDict):  # these are ports
    client: str
    dashboard: str
    metrics: str
    redis: str
    serve: str


class RayClusterHead(TypedDict):
    podIP: NotRequired[str]
    podName: NotRequired[str]
    serviceIP: NotRequired[str]
    serviceName: NotRequired[str]


class RayClusterStatus(TypedDict):
    desiredCPU: str
    desiredGPU: str
    desiredMemory: str
    desiredTPU: str

    lastUpdateTime: str  # "2024-09-16T09:34:29Z"
    maxWorkerReplicas: int
    observedGeneration: int

    head: NotRequired[RayClusterHead]
    endpoints: NotRequired[RayClusterEndpoints]
    state: NotRequired[str]
    conditions: NotRequired[list[dict[str, Any]]]


class RayClusterClient(BaseKubeRayClient[RayClusterStatus]):
    def __init__(
        self,
        kube_config: str | None = None,
        kube_context: str | None = None,
        api_client: ApiClient | None = None,
    ) -> None:
        self.kube_config = kube_config
        self.kube_context = kube_context

        # note: this call must happen BEFORE creating the api clients
        if api_client is None:
            load_kubeconfig(context=self.kube_context, config_file=self.kube_config)

        super().__init__(group=GROUP, version=VERSION, kind=KIND, plural=PLURAL, api_client=api_client)

    def wait_until_ready(
        self,
        name: str,
        namespace: str,
        timeout: float,
        failure_tolerance_timeout: float = 0.0,
        poll_interval: float = 5.0,
        log_cluster_conditions: bool = False,
    ) -> tuple[str, RayClusterEndpoints]:
        """
        If ready, returns service name (FQDN-ready) and a dictionary of ports.

        Parameters:
            name (str): The name of the `RayCluster` resource
            namespace (str): The namespace of the `RayCluster` resource
            timeout (float): The timeout in seconds to wait for the cluster to become ready.
            failure_tolerance_timeout (float): The period in seconds to wait for the cluster to transition out of `failed` state if it reaches it. This state can be transient under certain conditions. With the default value of 0, the first `failed` state appearance will raise an exception immediately.
            poll_interval (float): The interval in seconds to poll the cluster status.
            log_cluster_conditions (bool): Whether to log cluster conditions. See [KubeRay docs](https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/observability.html#raycluster-status-conditions)

        Returns:
            tuple[str, RayClusterEndpoints]: The service name (FQDN-ready) and a dictionary of ports.
        """
        start_time = time.time()

        condition_index_to_log = 0

        @retry(
            stop=stop_after_attempt(30),
            wait=wait_fixed(2),
            retry=(
                retry_if_exception_type(urllib3.exceptions.ProtocolError)
                # transient K8s API errors during cluster startup (e.g. 404 when the resource doesn't exist yet)
                | retry_if_exception(is_retryable_k8s_api_exception)
            ),
            reraise=True,
        )
        def get_status_with_retry() -> RayClusterStatus:
            return self.get_status(name=name, namespace=namespace, timeout=timeout)

        status = get_status_with_retry()

        while time.time() - start_time < timeout:
            status = get_status_with_retry()
            state = status.get("state")

            if not state:
                continue

            # https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/observability.html#raycluster-status-conditions
            conditions = list(reversed(status.get("conditions", [])))

            if log_cluster_conditions:
                while len(conditions) > condition_index_to_log:
                    logger.info(f"RayCluster {namespace}/{name} condition: {conditions[condition_index_to_log]}")
                    condition_index_to_log += 1

            if state == "failed" and (time.time() - start_time > failure_tolerance_timeout):
                error_msg = f"RayCluster {namespace}/{name} status transitioned into `failed`. Check the logs for more info. Consider increasing `failure_tolerance_timeout` ({failure_tolerance_timeout:.1f}s). Status:\n\n{status}\n\nMore details: `kubectl -n {namespace} describe RayCluster {name}`"

                logs = self._read_head_pod_logs(status=status, namespace=namespace, tail_lines=500)
                if logs:
                    logger.error(logs)

                raise RuntimeError(error_msg)

            if (
                state == "ready"
                and status.get("endpoints", {}).get("dashboard")
                and (head := status.get("head"))
                and head.get("serviceIP")
                and head.get("serviceName")
            ):
                logger.debug(f"RayCluster {namespace}/{name} is ready!")
                return head["serviceName"], status["endpoints"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

            time.sleep(poll_interval)
        else:
            # Try to fetch head pod logs for better debugging
            error_msg = f"Timed out ({timeout:.1f}s) waiting for RayCluster {namespace}/{name} to be ready. Check the logs for more info. Status: {status}"

            logs = self._read_head_pod_logs(status=status, namespace=namespace, tail_lines=500)
            if logs:
                logger.error(logs)

            raise TimeoutError(error_msg)

    @contextmanager
    def port_forward(
        self,
        name: str,
        namespace: str,
        local_dashboard_port: int = 8265,
        local_gcs_port: int = 10001,
    ) -> Iterator[tuple[int, int]]:
        """
        Port forwards the Ray dashboard and GCS ports to localhost.
        Use 0 for local_dashboard_port and local_gcs_port to get random available ports.
        Returns the ports that the dashboard and GCS are forwarded to.
        """
        if local_dashboard_port == 0:
            local_dashboard_port = get_random_available_port()

        if local_gcs_port == 0:
            local_gcs_port = get_random_available_port()

        service = f"{name}-head-svc"

        if not self.get_status(name, namespace):
            raise RuntimeError(f"RayCluster {name} does not exist in namespace {namespace}")

        self.wait_for_service_endpoints(service_name=service, namespace=namespace)

        cmd = [
            "kubectl",
            "port-forward",
            "-n",
            namespace,
            f"svc/{service}",
            f"{local_dashboard_port}:8265",
            f"{local_gcs_port}:10001",
        ]

        if self.kube_context:
            cmd.extend(["--context", self.kube_context])

        if self.kube_config:
            cmd.extend(["--kubeconfig", self.kube_config])

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        queue = Queue()
        SHOULD_STOP = False

        def should_stop():
            return SHOULD_STOP

        t = threading.Thread(target=enqueue_output, args=(process.stdout, queue, should_stop))
        t.daemon = True  # thread dies with the program
        t.start()

        time.sleep(0.01)

        try:
            # inspect stdout to avoid connecting too early
            # we must check for "Forwarding from" in stdout
            #

            while True:
                if process.poll() is not None:
                    raise RuntimeError(
                        f"port-forwarding command: `{' '.join(cmd)}` failed. Most likely ports {local_dashboard_port} or {local_gcs_port} are  already in use, or Kubernetes service {name}-head-svc does not exist in namespace {namespace}."
                    )

                line = queue.get()

                if "Forwarding from" in line:
                    break

            logger.info(
                f"Connecting to {namespace}/{name} via port-forwarding for ports {local_dashboard_port} and {local_gcs_port}..."
            )

            yield local_dashboard_port, local_gcs_port
        finally:
            # terminate the thread
            SHOULD_STOP = True
            process.kill()
            process.wait()
            t.join()
            logger.info(f"Port-forwarding for ports {local_dashboard_port} and {local_gcs_port} has been stopped.")

    @contextmanager
    def job_submission_client(
        self,
        name: str,
        namespace: str,
        port_forward: bool = False,
        timeout: float = 60,
        address: str | None = None,
        headers: dict[str, Any] | None = None,
        verify: str | bool = True,
        cookies: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Iterator[JobSubmissionClient]:
        """
        Returns a JobSubmissionClient object that can be used to interact with Ray jobs running in the KubeRay cluster.

        Args:
            name: Name of the RayCluster resource
            namespace: Namespace of the RayCluster resource
            port_forward: If True, port forward the dashboard to localhost
            timeout: Timeout for various operations
            address: Custom Ray dashboard address (e.g., "https://ray-cluster.example.com").
                When provided, connects directly to this address instead of using port-forwarding or in-cluster service IPs.
            headers: HTTP headers for dashboard authentication (e.g., {"Authorization": "Bearer token"}).
            verify: TLS certificate verification. Can be True, False, or path to CA bundle.
            cookies: Cookies to use when sending requests to the HTTP job server.
            metadata: Arbitrary metadata to store along with all jobs. Will be merged with per-job metadata.
        """

        from ray.job_submission import JobSubmissionClient

        if address:
            resolved_address = address
        elif port_forward:
            self.wait_for_service_endpoints(service_name=f"{name}-head-svc", namespace=namespace, timeout=timeout)
            with self.port_forward(name=name, namespace=namespace, local_dashboard_port=0, local_gcs_port=0) as (
                local_dashboard_port,
                _,
            ):
                resolved_address = f"http://localhost:{local_dashboard_port}"
                yield JobSubmissionClient(
                    address=resolved_address,
                    headers=headers,
                    verify=verify,
                    cookies=cookies,
                    metadata=metadata,
                )
            return
        else:
            # TODO: revisit the decision to use this as context-manager in this case
            status = self.get_status(name, namespace)
            service_name = status["head"]["serviceName"]  # type: ignore
            dashboard_port = status["endpoints"]["dashboard"]  # type: ignore
            resolved_address = f"http://{k8s_service_fqdn(service_name, namespace)}:{dashboard_port}"

        yield JobSubmissionClient(
            address=resolved_address,
            headers=headers,
            verify=verify,
            cookies=cookies,
            metadata=metadata,
        )

    def _read_head_pod_logs(self, status: RayClusterStatus, namespace: str, tail_lines: int = 500) -> str | None:
        if not ((head := status.get("head")) and (pod_name := head.get("podName"))):
            return None

        logs_output = []

        try:
            # First, get the pod to list all containers
            pod = self._core_v1_api.read_namespaced_pod(name=pod_name, namespace=namespace)

            if pod and pod.spec:  # pyright: ignore
                # Fetch logs from all containers except autoscaler
                for container in pod.spec.containers:  # pyright: ignore
                    container_name = container.name

                    # Skip autoscaler container - we only care about the main Ray head container
                    if container_name == "autoscaler":
                        continue

                    try:
                        container_logs = cast(
                            str,
                            self._core_v1_api.read_namespaced_pod_log(
                                name=pod_name,
                                namespace=namespace,
                                container=container_name,
                                tail_lines=tail_lines,
                            ),
                        )
                        logs_output.append(f"=== Container: {container_name} (last {tail_lines} lines) ===")
                        logs_output.append(container_logs)
                        logs_output.append("")  # Empty line between containers
                    except Exception:
                        logger.exception(
                            f"Failed to fetch logs for container {container_name} in pod {namespace}/{pod_name}"
                        )

            return "\n".join(logs_output) if logs_output else ""

        except Exception:
            logger.exception(f"Failed to fetch head pod {namespace}/{pod_name} metadata or logs")
            return None



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Literal

import dagster as dg
from pydantic import Field

USER_DEFINED_RAY_KEY = "dagster-ray/config"
DAGSTER_RAY_NAMESPACES_ENV_VAR = "DAGSTER_RAY_NAMESPACES"
DAGSTER_RAY_NAMESPACES_DEFAULT_VALUE = "ray"
DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_ENV_VAR = "DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS"
DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_DEFAULT_VALUE = str(4 * 60 * 60)


class Lifecycle(dg.Config):
    create: bool = Field(
        default=True,
        description="Whether to create the resource. If set to `False`, the user can manually call `.create` instead.",
    )
    wait: bool = Field(
        default=True,
        description="Whether to wait for the remote Ray cluster to become ready to accept connections. If set to `False`, the user can manually call `.wait` instead.",
    )
    connect: bool = Field(
        default=True,
        description="Whether to run `ray.init` against the remote Ray cluster. If set to `False`, the user can manually call `.connect` instead.",
    )
    cleanup: Literal["never", "always", "on_exception"] = Field(
        default="always",
        description="Resource cleanup policy. Determines when the resource should be deleted after Dagster step execution or during interruption.",
    )


class RayExecutionConfig(dg.Config):
    runtime_env: dict[str, Any] | None = Field(default=None, description="The runtime environment to use.")
    num_cpus: float | None = Field(default=None, description="The number of CPUs to allocate.")
    num_gpus: float | None = Field(default=None, description="The number of GPUs to allocate.")
    memory: int | None = Field(default=None, description="The amount of memory in bytes to allocate.")
    resources: dict[str, float] | None = Field(default=None, description="Custom resources to allocate.")

    @classmethod
    def from_tags(cls, tags: Mapping[str, str]) -> RayExecutionConfig:
        if USER_DEFINED_RAY_KEY in tags:
            return cls.model_validate_json(tags[USER_DEFINED_RAY_KEY])
        else:
            return cls()


class RayJobSubmissionClientConfig(dg.Config):
    address: str | None = Field(default=None, description="The address of the Ray cluster to connect to.")
    metadata: dict[str, str] | None = Field(
        default=None,
        description="""Arbitrary metadata to store along with all jobs. New metadata
            specified per job will be merged with the global metadata provided here
            via a simple dict update.""",
    )
    headers: dict[str, Any] | None = Field(
        default=None,
        description="""Headers to use when sending requests to the HTTP job server, used
            for cases like authentication to a remote cluster.""",
    )
    cookies: dict[str, Any] | None = Field(
        default=None, description="Cookies to use when sending requests to the HTTP job server."
    )


class ExecutionOptionsConfig(dg.Config):
    cpu: int | None = None
    gpu: int | None = None
    object_store_memory: int | None = None


class RayDataExecutionOptions(dg.Config):
    execution_options: ExecutionOptionsConfig = Field(default_factory=ExecutionOptionsConfig)

    cpu_limit: int = 5000
    gpu_limit: int = 0
    verbose_progress: bool = True
    use_polars: bool = True

    def apply(self):
        import ray
        from ray.data import ExecutionResources

        ctx = ray.data.DatasetContext.get_current()

        ctx.execution_options.resource_limits = ExecutionResources.for_limits(
            cpu=self.execution_options.cpu,
            gpu=self.execution_options.gpu,
            object_store_memory=self.execution_options.object_store_memory,
        )

        ctx.verbose_progress = self.verbose_progress
        ctx.use_polars = self.use_polars

    def apply_remote(self):
        import ray

        @ray.remote
        def apply():
            self.apply()

        ray.get(apply.remote())



from __future__ import annotations

import copy
import json
import time
import warnings
from typing import TYPE_CHECKING, Any, cast

import dagster as dg
import dagster._check as check
import yaml
from dagster import (
    PipesContextInjector,
    PipesEnvContextInjector,
    PipesMessageReader,
    PipesSession,
    open_pipes_session,
)
from dagster._core.definitions.resource_annotation import TreatAsResourceParam
from dagster._core.errors import DagsterExecutionInterruptedError
from dagster._core.pipes.client import PipesClientCompletedInvocation
from dagster_pipes import PipesExtras

from dagster_ray._base.utils import get_dagster_tags
from dagster_ray.core.pipes import (
    PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY,
    PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY,
    PipesRayJobMessageReader,
    SubmitJobParams,
    generate_job_id,
)
from dagster_ray.kuberay.client import RayJobClient
from dagster_ray.kuberay.client.rayjob.client import RayJobStatus
from dagster_ray.kuberay.configs import RayJobConfig, RayJobSpec
from dagster_ray.kuberay.utils import k8s_service_fqdn, normalize_k8s_label_values
from dagster_ray.types import OpOrAssetExecutionContext

if TYPE_CHECKING:
    from ray.job_submission import JobSubmissionClient


def _build_ray_resources_param(resources: dict[str, float]) -> str:
    return '"' + json.dumps(resources).replace('"', '\\"').replace("'", '\\"') + '"'


def _ray_job_from_submit_params(
    context: OpOrAssetExecutionContext,
    submit_job_params: SubmitJobParams,
    namespace: str = "ray",
) -> dict[str, Any]:
    """Build a minimal KubeRay RayJob manifest from submit_job_params.

    Reuses ``RayJobConfig`` / ``RayJobSpec`` to generate the manifest,
    consistent with how ``KubeRayCluster`` and ``KubeRayInteractiveJob``
    auto-generate their specs. Uses the ``dagster/image`` run tag for
    the container image.
    """
    runtime_env_yaml = None
    if runtime_env := submit_job_params.get("runtime_env"):
        runtime_env_yaml = yaml.safe_dump(runtime_env, default_style='"')

    ray_job_config = RayJobConfig(
        metadata={"namespace": namespace},
        spec=RayJobSpec(
            runtime_env_yaml=runtime_env_yaml,
            entrypoint_num_cpus=submit_job_params.get("entrypoint_num_cpus"),
            entrypoint_num_gpus=submit_job_params.get("entrypoint_num_gpus"),
            entrypoint_memory=submit_job_params.get("entrypoint_memory"),
            entrypoint_resources=_build_ray_resources_param(er)
            if (er := submit_job_params.get("entrypoint_resources")) is not None
            else None,
            metadata=submit_job_params.get("metadata"),
            job_id=submit_job_params.get("submission_id"),
        ),
    )

    image = context.run.tags.get("dagster/image")
    ray_job = ray_job_config.to_k8s(context=context, image=image)
    ray_job["metadata"]["namespace"] = namespace
    ray_job["spec"]["entrypoint"] = submit_job_params["entrypoint"]

    return ray_job


def _merge_submit_params_into_ray_job(
    ray_job: dict[str, Any],
    submit_job_params: SubmitJobParams,
) -> dict[str, Any]:
    """Merge ``submit_job_params`` into a KubeRay RayJob manifest, warning on conflicts."""
    ray_job = copy.deepcopy(ray_job)

    def _set(key: str, value: Any) -> None:
        if key in ray_job["spec"] and ray_job["spec"][key] != value:
            warnings.warn(
                f"submit_job_params overwriting ray_job spec field '{key}': {ray_job['spec'][key]!r} -> {value!r}",
                stacklevel=3,
            )
        ray_job["spec"][key] = value

    _set("entrypoint", submit_job_params["entrypoint"])

    if runtime_env := submit_job_params.get("runtime_env"):
        _set("runtimeEnvYAML", yaml.safe_dump(runtime_env, default_style='"'))

    if (num_cpus := submit_job_params.get("entrypoint_num_cpus")) is not None:
        _set("entrypointNumCpus", num_cpus)

    if (num_gpus := submit_job_params.get("entrypoint_num_gpus")) is not None:
        _set("entrypointNumGpus", num_gpus)

    if (memory := submit_job_params.get("entrypoint_memory")) is not None:
        _set("entrypointMemory", memory)

    if (resources := submit_job_params.get("entrypoint_resources")) is not None:
        _set("entrypointResources", _build_ray_resources_param(resources))

    if (metadata := submit_job_params.get("metadata")) is not None:
        _set("metadata", metadata)

    if (submission_id := submit_job_params.get("submission_id")) is not None:
        _set("jobId", submission_id)

    return ray_job


class PipesKubeRayJobClient(dg.PipesClient, TreatAsResourceParam):
    """A pipes client for running ``RayJob`` on Kubernetes.

    Most constructor arguments can be overridden in the [`run`][dagster_ray.kuberay.pipes.PipesKubeRayJobClient.run] method.

    Args:
        context_injector: A context injector to use to inject
            context into the ``RayJob``. Defaults to [`PipesEnvContextInjector`][dagster.PipesEnvContextInjector].
        message_reader: A message reader to use to read messages
            from the glue job run. Defaults to [`PipesRayJobMessageReader`][dagster_ray.core.pipes.PipesRayJobMessageReader].
        client: The Kubernetes API client.
        forward_termination: Whether to terminate the Ray job when the Dagster process receives a termination signal,
            or if the startup timeout is reached. Defaults to ``True``.
        timeout: Timeout for various internal interactions with the Kubernetes RayJob.
        poll_interval: Interval at which to poll Kubernetes for status updates.
        port_forward: Whether to use Kubernetes port-forwarding to connect to the KubeRay cluster.
            Is useful when running in a local environment.
        address: Ray dashboard address. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
            Generated by [`resolve_hostname`][dagster_ray.kuberay.pipes.PipesKubeRayJobClient.resolve_hostname] if not provided.
        headers: HTTP headers for Ray Dashboard requests. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        verify: Whether to verify TLS certificate. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        cookies: HTTP cookies for Ray Dashboard requests. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        metadata: Ray Job metadata. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].

    Info:
        Image defaults to `dagster/image` run tag.

    Tip:
        Make sure `ray[full]` is available in the image.
    """

    def __init__(
        self,
        client: RayJobClient | None = None,
        context_injector: PipesContextInjector | None = None,
        message_reader: PipesMessageReader | None = None,
        forward_termination: bool = True,
        timeout: float = 600,
        poll_interval: float = 1,
        port_forward: bool = False,
        address: str | None = None,
        headers: dict[str, Any] | None = None,
        verify: str | bool = True,
        cookies: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ):
        self.client: RayJobClient = client or RayJobClient()

        self._context_injector = context_injector or PipesEnvContextInjector()

        self._message_reader = message_reader or PipesRayJobMessageReader(
            job_submission_client_kwargs={
                "headers": headers,
                "verify": verify,
                "cookies": cookies,
                "metadata": metadata,
            }
        )

        self.forward_termination = check.bool_param(forward_termination, "forward_termination")
        self.timeout = check.numeric_param(timeout, "timeout")
        self.poll_interval = check.numeric_param(poll_interval, "poll_interval")
        self.port_forward = check.bool_param(port_forward, "port_forward")
        self.address = address
        self.headers = headers
        self.verify = verify
        self.cookies = cookies
        self.metadata = metadata

        self._job_submission_client: JobSubmissionClient | None = None

    @property
    def job_submission_client(self) -> JobSubmissionClient:
        if self._job_submission_client is None:
            raise dg.DagsterInvariantViolationError("JobSubmissionClient is only available inside the run method.")
        else:
            return self._job_submission_client

    def run(  # type: ignore
        self,
        *,
        context: OpOrAssetExecutionContext,
        submit_job_params: SubmitJobParams | None = None,
        ray_job: dict[str, Any] | None = None,
        extras: PipesExtras | None = None,
        address: str | None = None,
        headers: dict[str, Any] | None = None,
        verify: str | bool | None = None,
        cookies: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> PipesClientCompletedInvocation:
        """Execute a RayJob, enriched with the Pipes protocol.

        Parameters passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient] take precedence over the constructor arguments.

        At least one of ``submit_job_params`` or ``ray_job`` must be provided.

        Args:
            context: Current Dagster op or asset context.
            submit_job_params: Job submission parameters (entrypoint, runtime_env, etc.). When combined with ``ray_job``, values are merged into the template.  When used alone, a minimal manifest is auto-generated using the ``dagster/image`` run tag.
            ray_job: KubeRay RayJob template dict. Can be used alone or together with ``submit_job_params``.
            extras: Additional information to pass to the Pipes session, retrievable via [`PipesContext.get_extras`][dagster_pipes.PipesContext.get_extra].
            address: Ray dashboard address. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
                Generated by [`resolve_hostname`][dagster_ray.kuberay.pipes.PipesKubeRayJobClient.resolve_hostname] if not provided.
            headers: HTTP headers for Ray Dashboard requests. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
            verify: Whether to verify TLS certificate. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
            cookies: HTTP cookies for Ray Dashboard requests. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
            metadata: Ray Job metadata. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        """
        if submit_job_params is None and ray_job is None:
            raise dg.DagsterInvariantViolationError(
                "At least one of `submit_job_params` or `ray_job` must be provided."
            )

        if ray_job is not None and submit_job_params is not None:
            ray_job = _merge_submit_params_into_ray_job(ray_job, submit_job_params)
        elif ray_job is not None:
            ray_job = copy.deepcopy(ray_job)
        else:
            ray_job = _ray_job_from_submit_params(context, submit_job_params)  # type: ignore[arg-type]

        with open_pipes_session(
            context=context,
            message_reader=self._message_reader,
            context_injector=self._context_injector,
            extras=extras,
        ) as session:
            ray_job = self._enrich_ray_job(context, session, ray_job)
            start_response = self._start(context, session, ray_job)
            start_status = cast(RayJobStatus, start_response["status"])
            ray_job_id = start_status["jobId"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

            name = ray_job["metadata"]["name"]
            namespace = ray_job["metadata"]["namespace"]

            cluster_name = self.client.get_ray_cluster_name(
                name=name, namespace=namespace, timeout=self.timeout, poll_interval=self.poll_interval
            )

            resolved_address = address or self.address
            if resolved_address is None and not self.port_forward:
                cluster_status = self.client.ray_cluster_client.get_status(
                    name=cluster_name, namespace=namespace, timeout=self.timeout, poll_interval=self.poll_interval
                )
                service_name = cluster_status["head"]["serviceName"]  # type: ignore
                dashboard_port = int(cluster_status["endpoints"]["dashboard"])  # type: ignore
                resolved_address = self.resolve_hostname(service_name, namespace, dashboard_port)

            with self.client.ray_cluster_client.job_submission_client(
                name=cluster_name,
                namespace=namespace,
                port_forward=self.port_forward,
                address=resolved_address,
                headers=headers or self.headers,
                verify=verify if verify is not None else self.verify,
                cookies=cookies or self.cookies,
                metadata=metadata or self.metadata,
            ) as job_submission_client:
                self._job_submission_client = job_submission_client

                session.report_launched(
                    {
                        "extras": {
                            PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY: ray_job_id,
                            PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY: job_submission_client.get_address(),
                        }
                    }
                )

                try:
                    self._wait_for_completion(context, start_response)

                    if isinstance(self._message_reader, PipesRayJobMessageReader) and self.port_forward:
                        # in this case the message reader will fail once port forwarding is finished
                        # TODO: merge https://github.com/danielgafni/dagster-ray/pull/123
                        # to avoid this work-around
                        self._message_reader.thread_ready.wait()
                        context.log.debug(
                            "[pipes] waiting for PipesRayJobMessageReader to complete before stopping port-forwarding"
                        )
                        self._message_reader.session_closed.set()
                        self._message_reader.completed.wait()

                    return PipesClientCompletedInvocation(
                        session, metadata={"RayJob": f"{namespace}/{name}", "Ray Job ID": ray_job_id}
                    )

                except DagsterExecutionInterruptedError:
                    if self.forward_termination:
                        context.log.warning(
                            f"[pipes] Dagster process interrupted! Will terminate RayJob {namespace}/{name}."
                        )
                        self._terminate(
                            context,
                            start_response,
                            address=resolved_address,
                            headers=headers,
                            verify=verify,
                            cookies=cookies,
                            metadata=metadata,
                        )
                    raise

    def resolve_hostname(self, service_name: str, namespace: str, dashboard_port: int) -> str:
        """Resolve the Ray dashboard address from a Kubernetes service name, namespace, and port.

        By default, returns an HTTP URL using the in-cluster FQDN
        (e.g. `http://my-svc.namespace.svc.cluster.local:8265`).

        Override this method to customize the address, for example to use a proxy
        or a different protocol. The `dashboard_port` can be ignored if not needed.

        Args:
            service_name: The Kubernetes service name for the Ray head node.
            namespace: The Kubernetes namespace.
            dashboard_port: The Ray dashboard port reported by the KubeRay operator.

        Returns:
            The full Ray dashboard URL (including protocol and optional port).

        Example:
            ```python
            class MyPipesKubeRayJobClient(PipesKubeRayJobClient):
                def resolve_hostname(self, service_name: str, namespace: str, dashboard_port: int) -> str:
                    return f"https://{service_name}.{namespace}.company.com"
            ```
        """
        hostname = k8s_service_fqdn(service_name, namespace)
        return f"http://{hostname}:{dashboard_port}"

    def get_dagster_tags(self, context: OpOrAssetExecutionContext) -> dict[str, str]:
        tags = get_dagster_tags(context)
        return tags

    def _enrich_ray_job(
        self, context: OpOrAssetExecutionContext, session: PipesSession, ray_job: dict[str, Any]
    ) -> dict[str, Any]:
        env_vars = session.get_bootstrap_env_vars()

        ray_job["metadata"] = ray_job.get("metadata", {})
        ray_job["metadata"]["labels"] = ray_job["metadata"].get("labels", {})

        ray_job["metadata"]["name"] = ray_job["metadata"].get("name", f"pipes-{generate_job_id()}")
        ray_job["metadata"]["labels"].update(normalize_k8s_label_values(self.get_dagster_tags(context)))

        # update env vars in runtimeEnv
        runtime_env_yaml = ray_job["spec"].get("runtimeEnvYAML", "{}")

        runtime_env = yaml.safe_load(runtime_env_yaml)
        runtime_env["env_vars"] = runtime_env.get("env_vars", {})
        runtime_env["env_vars"].update(env_vars)

        ray_job["spec"]["runtimeEnvYAML"] = yaml.safe_dump(runtime_env, default_style='"')

        image_from_run_tag = context.run.tags.get("dagster/image")

        for container in ray_job["spec"]["rayClusterSpec"]["headGroupSpec"]["template"]["spec"]["containers"]:
            container["image"] = container.get("image") or image_from_run_tag

        for worker_spec in ray_job["spec"]["rayClusterSpec"]["workerGroupSpecs"]:
            for container in worker_spec["template"]["spec"]["containers"]:
                container["image"] = container.get("image") or image_from_run_tag

        return ray_job

    def _start(
        self, context: OpOrAssetExecutionContext, session: PipesSession, ray_job: dict[str, Any]
    ) -> dict[str, Any]:
        name = ray_job["metadata"]["name"]
        namespace = ray_job["metadata"]["namespace"]

        context.log.info(f"[pipes] Starting RayJob {namespace}/{name}...")

        self.client.create(
            body=ray_job,
            namespace=ray_job["metadata"]["namespace"],
        )

        self.client.wait_until_running(
            name=name,
            namespace=namespace,
            timeout=self.timeout,
            poll_interval=self.poll_interval,
            terminate_on_timeout=self.forward_termination,
            port_forward=self.port_forward,
        )

        return self.client.get(
            namespace=ray_job["metadata"]["namespace"],
            name=ray_job["metadata"]["name"],
        )

    def _wait_for_completion(self, context: OpOrAssetExecutionContext, start_response: dict[str, Any]) -> RayJobStatus:
        context.log.info("[pipes] Waiting for RayJob to complete...")

        name = start_response["metadata"]["name"]
        namespace = start_response["metadata"]["namespace"]

        while True:
            status = self.client.get_status(
                name=name,
                namespace=namespace,
            )

            if job_status := status.get("jobStatus"):
                if job_status in ["PENDING", "RUNNING"]:
                    pass
                elif job_status == "SUCCEEDED":
                    context.log.info(f"[pipes] RayJob {namespace}/{name} succeeded!")
                    return status
                elif job_status in ["STOPPED", "FAILED"]:
                    raise RuntimeError(
                        f"RayJob {namespace}/{name} status is {job_status}. Message:\n{status.get('message')}"
                    )
                else:
                    raise RuntimeError(
                        f"RayJob {namespace}/{name} has an unknown status: {job_status}. Message:\n{status.get('message')}"
                    )

            time.sleep(self.poll_interval)

    def _terminate(
        self,
        context: OpOrAssetExecutionContext,
        start_response: dict[str, Any],
        address: str | None = None,
        headers: dict[str, Any] | None = None,
        verify: str | bool | None = None,
        cookies: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        name = start_response["metadata"]["name"]
        namespace = start_response["metadata"]["namespace"]

        context.log.info(f"[pipes] Terminating RayJob {namespace}/{name} ...")
        self.client.terminate(
            name=name,
            namespace=namespace,
            port_forward=self.port_forward,
            address=address,
            headers=headers or self.headers,
            verify=verify if verify is not None else self.verify,
            cookies=cookies or self.cookies,
            metadata=metadata or self.metadata,
        )
        context.log.info(f"[pipes] RayJob {namespace}/{name} terminated.")



from collections.abc import Sequence
from datetime import datetime
from typing import cast

import dagster as dg
from pydantic import Field
from typing_extensions import override

from dagster_ray._base.cluster_sharing_lock import ClusterSharingLock
from dagster_ray.configs import Lifecycle
from dagster_ray.kuberay.client import RayClusterClient
from dagster_ray.kuberay.configs import ClusterSharing, RayClusterConfig
from dagster_ray.kuberay.leader_election import LeaderElection
from dagster_ray.kuberay.resources.base import BaseKubeRayResource
from dagster_ray.kuberay.utils import normalize_k8s_label_values
from dagster_ray.types import AnyDagsterContext
from dagster_ray.utils import get_dagster_run


class KubeRayClusterClientResource(dg.ConfigurableResource[RayClusterClient]):
    """This configurable resource provides a [dagster_ray.kuberay.client.RayClusterClient][]."""

    kube_context: str | None = None
    kube_config: str | None = None

    def create_resource(self, context: dg.InitResourceContext) -> RayClusterClient:
        client = RayClusterClient(
            kube_context=self.kube_context,
            kube_config=self.kube_config,
        )
        return client


class KubeRayCluster(BaseKubeRayResource):
    """
    Provides a Ray Cluster for Dagster steps.

    It is advised to use [`KubeRayInteractiveJob`][dagster_ray.kuberay.resources.KubeRayInteractiveJob] with KubeRay >= 1.3.0 instead.

    Info:
        Image defaults to `dagster/image` run tag.

    Tip:
        Make sure `ray[full]` is available in the image.
    """

    lifecycle: Lifecycle = Field(
        default_factory=lambda: Lifecycle(cleanup="always"), description="Actions to perform during resource setup."
    )

    cluster_sharing: ClusterSharing = Field(
        default_factory=ClusterSharing,
        description="Configuration for sharing the `RayCluster` across Dagster steps. Existing clusters matching this configuration will be reused without recreating them. A `dagster/sharing=true` label will be applied to the `RayCluster`, and a `dagster/lock-<run-id>-<step-id>=<lock>` annotation will be placed on the `RayCluster` to mark it as being used by this step. Cleanup will only proceed if the `RayCluster` is not being used by any other steps, therefore cluster sharing should be used in conjunction with [dagster_ray.kuberay.sensors.cleanup_expired_kuberay_clusters][] sensor.",
    )

    ray_cluster: RayClusterConfig = Field(
        default_factory=RayClusterConfig, description="Kubernetes `RayCluster` CR configuration."
    )
    client: dg.ResourceDependency[RayClusterClient] = Field(  # pyright: ignore[reportAssignmentType]
        default_factory=KubeRayClusterClientResource, description="Kubernetes `RayCluster` client"
    )
    log_cluster_conditions: bool = Field(
        default=True,
        description="Whether to log RayCluster conditions while waiting for the RayCluster to become ready. Learn more: [KubeRay docs](https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/observability.html#raycluster-status-conditions).",
    )

    @property
    def host(self) -> str:
        if self._host is None:
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._host

    @property
    def name(self) -> str:
        if (name := self.ray_cluster.metadata.get("name")) is not None:
            return name
        if self._name is None:
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._name

    @property
    def namespace(self) -> str:
        return self.ray_cluster.namespace

    @property
    @override
    def display_name(self) -> str:
        return (
            f"RayCluster {self.namespace}/{self.name}" if self.created else f"RayCluster in namespace {self.namespace}"
        )

    def get_dagster_tags(self, context: AnyDagsterContext) -> dict[str, str]:
        tags = super().get_dagster_tags(context=context)
        tags.update({"dagster/deployment": self.deployment_name})
        if self.cluster_sharing.enabled:
            tags.update({"dagster/cluster-sharing": "true"})
        return tags

    def get_k8s_labels(self, context: AnyDagsterContext) -> dict[str, str]:
        return normalize_k8s_label_values(self.get_dagster_tags(context))

    def get_image(self, context: AnyDagsterContext) -> str | None:
        assert context.dagster_run is not None
        return self.image or context.dagster_run.tags.get("dagster/image")

    def _find_shared_cluster(self, label_selector: str) -> list[dict]:
        """Find existing shared clusters matching the label selector."""
        return self.client.list(
            label_selector=label_selector,
            namespace=self.namespace,
        ).get("items", [])

    def _use_existing_cluster(self, context: AnyDagsterContext, cluster_name: str) -> None:
        """Mark an existing cluster as being used by this step."""
        self._name = cluster_name
        self._creation_verb = "Using"

        # Place a lock on the cluster using JSON patch to avoid overwriting existing annotations
        lock_annotations = self.get_sharing_lock_annotations(context)
        patch_operations = [
            {
                "op": "add",
                "path": f"/metadata/annotations/{key.replace('/', '~1')}",
                "value": value,
            }
            for key, value in lock_annotations.items()
        ]

        self.client.update_json_patch(
            name=cluster_name,
            namespace=self.namespace,
            body=patch_operations,
        )

    def _get_lease_name(self, context: AnyDagsterContext) -> str:
        """Generate a deterministic lease name for cluster sharing coordination.

        The lease name must be the same for all steps that should share a cluster,
        so it's derived from the sharing label selector (which defines "same cluster" semantics).
        The label selector is sorted to ensure deterministic ordering.
        """
        import hashlib

        label_selector = self.get_sharing_label_selector(context)
        # Sort the label selector components for deterministic ordering
        # Label selector format is "key1=value1,key2=value2,..."
        sorted_selector = ",".join(sorted(label_selector.split(",")))
        # Create a short hash of the sorted label selector for the lease name
        selector_hash = hashlib.sha256(sorted_selector.encode()).hexdigest()[:12]
        return f"dagster-ray-{selector_hash}"

    @override
    def create(self, context: AnyDagsterContext):
        assert context.log is not None
        assert context.dagster_run is not None

        labels = self.get_k8s_labels(context)
        annotations: dict[str, str] = {}

        if self.cluster_sharing.enabled:
            label_selector = self.get_sharing_label_selector(context)
            context.log.info(
                f"RayCluster sharing is enabled. Looking for clusters matching label selector: {label_selector}"
            )

            # Check if a shared cluster already exists
            matching_clusters = self._find_shared_cluster(label_selector)
            if matching_clusters:
                cluster_name = matching_clusters[0]["metadata"]["name"]
                context.log.info(
                    f"Found {len(matching_clusters)} clusters matching the label selector. "
                    f"Using the first one: {cluster_name}"
                )
                self._use_existing_cluster(context, cluster_name)
                return

            # No cluster exists - use leader election to coordinate creation
            context.log.info("No matching shared clusters found. Using leader election to coordinate cluster creation.")

            lease_name = self._get_lease_name(context)
            holder_identity = f"{context.run_id}-{self.resource_uid}"
            leader_election = LeaderElection(
                holder_identity=holder_identity,
                api_client=self.client.api_client,
                lease_duration_seconds=10,
                retry_period_seconds=2.0,
                acquire_timeout_seconds=60.0,
            )

            def cluster_exists() -> bool:
                return bool(self._find_shared_cluster(label_selector))

            result = leader_election.acquire_or_wait(
                lease_name=lease_name,
                namespace=self.namespace,
                resource_exists_check=cluster_exists,
            )

            if not result.is_leader:
                # Another step created the cluster while we were waiting
                matching_clusters = self._find_shared_cluster(label_selector)
                if matching_clusters:
                    cluster_name = matching_clusters[0]["metadata"]["name"]
                    context.log.info(f"Using cluster created by leader: {cluster_name}")
                    self._use_existing_cluster(context, cluster_name)
                    return
                else:
                    # Edge case: leader created then deleted? Fall through to create
                    context.log.warning(
                        "Leader election indicated not leader, but no cluster found. Proceeding to create cluster."
                    )

            context.log.info("Won leader election - this step will create the shared cluster.")

            # Mark the cluster as being used by this step
            annotations.update(self.get_sharing_lock_annotations(context))

            # Release the lease after we're done creating (cluster existence is the coordination point now)
            # We'll release after successful creation below

        self._name = self.ray_cluster.metadata.get("name") or self._get_step_name(context)

        # Safety measure: don't recreate the cluster for step retries
        if not self.client.get(
            name=self.name,
            namespace=self.namespace,
        ):
            k8s_manifest = self.ray_cluster.to_k8s(
                context,
                image=self.get_image(context),
                labels=labels,
                annotations=annotations,
                env_vars=self.get_env_vars_to_inject(),
            )

            k8s_manifest["metadata"]["name"] = self.name

            resource = self.client.create(body=k8s_manifest, namespace=self.namespace)
            if not resource:
                raise RuntimeError(f"Couldn't create {self.display_name}")

            # Release the lease now that the cluster is created
            if self.cluster_sharing.enabled:
                try:
                    lease_name = self._get_lease_name(context)
                    holder_identity = f"{context.run_id}-{self.resource_uid}"
                    leader_election = LeaderElection(
                        holder_identity=holder_identity,
                        api_client=self.client.api_client,
                    )
                    leader_election.release(lease_name, self.namespace)
                except Exception as e:
                    # Non-fatal: lease will expire on its own
                    context.log.debug(f"Failed to release leader election lease: {e}")

    @override
    def wait(self, context: AnyDagsterContext):
        assert context.log is not None
        assert context.dagster_run is not None

        self.client.wait_until_ready(
            self.name,
            namespace=self.namespace,
            timeout=self.timeout,
            failure_tolerance_timeout=self.failure_tolerance_timeout,
            poll_interval=self.poll_interval,
            log_cluster_conditions=self.log_cluster_conditions,
        )

        service_name = self.client.get_status(
            name=self.name, namespace=self.namespace, timeout=self.timeout, poll_interval=self.poll_interval
        )[  # pyright: ignore
            "head"
        ]["serviceName"]
        self._host = self.resolve_hostname(service_name, self.namespace)

    @override
    def on_ready(self, context: AnyDagsterContext):
        assert context.log is not None
        msg = f"RayCluster {self.namespace}/{self.name} is ready! Connection command:\n"
        msg += f"kubectl -n {self.namespace} port-forward svc/{self.name}-head-svc 8265:8265 6379:6379 10001:10001"
        context.log.info(msg)

    @override
    def delete(self, context: AnyDagsterContext):
        self.client.delete(self.name, namespace=self.namespace)

    @override
    def cleanup(self, context: AnyDagsterContext, exception: BaseException | None):
        assert context.log is not None
        assert context.run_id is not None
        # we don't want to perform cleanup if:
        # - cluster sharing is enabled
        # - cluster has at least one lock that hasn't expired yet
        if self.cluster_sharing.enabled:
            # get sharing locks created by this or another Dagster step
            alive_locks = self.get_cluster_sharing_alive_locks(context)
            if len(alive_locks) > 0:
                context.log.info(
                    f"Skipping cluster cleanup due to active cluster sharing locks: {', '.join([lock.identifier for lock in alive_locks])}"
                )
                if self.connected and self._context is not None:
                    self._context.disconnect()
                self.on_cleanup(context, deleted=False)
                return

        super().cleanup(context, exception)

    def get_sharing_label_selector(self, context: AnyDagsterContext) -> str:
        """This method combines user-provided label selectors from the sharing config with default (dagster-generated) labels to match on.

        User-provided labels take priority. This method can be overridden to customize cluster sharing behavior.
        """
        labels = self.get_k8s_labels(context)

        dagster_match_labels: dict[str, str] = {}

        if self.cluster_sharing:
            if self.cluster_sharing.match_dagster_labels.cluster_sharing:
                dagster_match_labels["dagster/cluster-sharing"] = "true"
            if self.cluster_sharing.match_dagster_labels.code_location and (
                code_location := labels.get("dagster/code-location")
            ):
                dagster_match_labels["dagster/code-location"] = code_location
            if self.cluster_sharing.match_dagster_labels.resource_key:
                dagster_match_labels["dagster/resource-key"] = labels["dagster/resource-key"]
            if self.cluster_sharing.match_dagster_labels.run_id:
                dagster_match_labels["dagster/run-id"] = labels["dagster/run-id"]
            if self.cluster_sharing.match_dagster_labels.git_sha and (git_sha := labels.get("dagster/git-sha")):
                dagster_match_labels["dagster/git-sha"] = git_sha

        combined_match_labels = {
            **dagster_match_labels,
            **(self.cluster_sharing.match_labels or {}),
        }

        return ",".join([f"{key}={value}" for key, value in combined_match_labels.items()])

    def get_sharing_lock_annotations(self, context: AnyDagsterContext) -> dict[str, str]:
        run = get_dagster_run(context)
        lock = ClusterSharingLock(
            run_id=run.run_id,
            key=self.resource_uid,
            ttl_seconds=self.cluster_sharing.ttl_seconds,
            created_at=datetime.now(),
        )
        return {lock.tag: lock.model_dump_json()}

    def get_cluster_sharing_alive_locks(self, context: AnyDagsterContext) -> Sequence[ClusterSharingLock]:
        locks = ClusterSharingLock.parse_all_locks(
            cast(
                dict[str, str],
                self.client.get(name=self.name, namespace=self.namespace).get("metadata", {}).get("annotations", {}),
            )
        )
        return ClusterSharingLock.get_alive_locks(locks)



import os
from collections.abc import Mapping
from typing import Any, Literal

import dagster as dg
from pydantic import Field

from dagster_ray.kuberay.utils import remove_none_from_dict
from dagster_ray.types import AnyDagsterContext

in_k8s = os.environ.get("KUBERNETES_SERVICE_HOST") is not None


DEFAULT_AUTOSCALER_OPTIONS = {
    "upscalingMode": "Default",
    "idleTimeoutSeconds": 60,
    "env": [],
    "envFrom": [],
    "resources": {
        "limits": {"cpu": "50m", "memory": "0.1Gi"},
        "requests": {"cpu": "50m", "memory": "0.1Gi"},
    },
}
DEFAULT_HEAD_GROUP_SPEC = {
    "serviceType": "ClusterIP",
    "rayStartParams": {},
    "metadata": {
        "labels": {},
        "annotations": {},
    },
    "template": {
        "spec": {
            "imagePullSecrets": [],
            "containers": [
                {
                    "volumeMounts": [
                        {"mountPath": "/tmp/ray", "name": "ray-logs"},
                    ],
                    "name": "head",
                    "imagePullPolicy": "Always",
                },
            ],
            "volumes": [
                {"name": "ray-logs", "emptyDir": {}},
            ],
            "affinity": {},
            "tolerations": [],
            "nodeSelector": {},
        },
    },
}
DEFAULT_WORKER_GROUP_SPECS = [
    {
        "groupName": "workers",
        "replicas": 0,
        "minReplicas": 0,
        "maxReplicas": 1,
        "rayStartParams": {},
        "template": {
            "metadata": {"labels": {}, "annotations": {}},
            "spec": {
                "imagePullSecrets": [],
                "containers": [
                    {
                        "volumeMounts": [{"mountPath": "/tmp/ray", "name": "ray-logs"}],
                        "name": "worker",
                        "imagePullPolicy": "Always",
                    }
                ],
                "volumes": [
                    {"name": "ray-logs", "emptyDir": {}},
                ],
                "affinity": {},
                "tolerations": [],
                "nodeSelector": {},
            },
        },
    }
]


MISSING_IMAGE_MESSAGE = "Image is missing from the `RayCluster` spec, from the top-level Dagster resource config, and the Dagster run does not have a `dagster/image` tag. Please use one of these options to specify the image."


class AuthOptions(dg.Config):
    """[AuthOptions](https://ray-project.github.io/kuberay/reference/api/#authoptions) for the Ray cluster."""

    mode: Literal["token", "disabled"] = "token"


class RayClusterSpec(dg.PermissiveConfig):
    """[RayCluster spec](https://ray-project.github.io/kuberay/reference/api/#rayclusterspec) configuration options. A few sensible defaults are provided for convenience."""

    suspend: bool | None = None
    managed_by: str | None = None
    autoscaler_options: dict[str, Any] = DEFAULT_AUTOSCALER_OPTIONS
    head_service_annotations: dict[str, str] | None = None
    enable_in_tree_autoscaling: bool = False
    gcs_fault_tolerance_options: dict[str, Any] | None = None
    head_group_spec: dict[str, Any] = DEFAULT_HEAD_GROUP_SPEC
    ray_version: str | None = None
    worker_group_specs: list[dict[str, Any]] = DEFAULT_WORKER_GROUP_SPECS
    auth_options: AuthOptions | None = None

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""

        assert context.log is not None

        # TODO: inject self.redis_port and self.dashboard_port into the RayCluster configuration
        # TODO: auto-apply some tags from dagster-k8s/config

        head_group_spec = self.head_group_spec.copy()
        worker_group_specs = self.worker_group_specs.copy()

        k8s_env_vars: list[dict[str, Any]] = []

        if env_vars:
            for key, value in env_vars.items():
                k8s_env_vars.append({"name": key, "value": value})

        def update_group_spec(group_spec: dict[str, Any]):
            # TODO: only inject if the container has a `dagster.io/inject-image` annotation or smth
            if group_spec["template"]["spec"]["containers"][0].get("image") is None:
                if image is None:
                    raise ValueError(MISSING_IMAGE_MESSAGE)
                else:
                    group_spec["template"]["spec"]["containers"][0]["image"] = image

            for container in group_spec["template"]["spec"]["containers"]:
                container["env"] = container.get("env", []) + k8s_env_vars

        update_group_spec(head_group_spec)
        for worker_group_spec in worker_group_specs:
            update_group_spec(worker_group_spec)

        return remove_none_from_dict(
            {
                "enableInTreeAutoscaling": self.enable_in_tree_autoscaling,
                "autoscalerOptions": self.autoscaler_options,
                "headGroupSpec": head_group_spec,
                "workerGroupSpecs": worker_group_specs,
                "suspend": self.suspend,
                "managedBy": self.managed_by,
                "headServiceAnnotations": self.head_service_annotations,
                "gcsFaultToleranceOptions": self.gcs_fault_tolerance_options,
                "rayVersion": self.ray_version,
                "authOptions": self.auth_options,
            }
        )


class RayClusterConfig(dg.Config):
    kind: str = "RayCluster"
    api_version: str = "ray.io/v1"
    metadata: dict[str, Any] = Field(
        default_factory=dict,
        description="Kubernetes metadata, except the name field can be omitted. In this case it will be generated by `dagster-ray`.",
    )
    spec: RayClusterSpec = Field(default_factory=RayClusterSpec)

    @property
    def namespace(self) -> str:
        return self.metadata.get("namespace", "default")

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        labels: Mapping[str, str] | None = None,
        annotations: Mapping[str, str] | None = None,
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        assert context.log is not None
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""

        labels = labels or {}
        annotations = annotations or {}

        return {
            "apiVersion": self.api_version,
            "kind": self.kind,
            "metadata": remove_none_from_dict(
                {
                    "name": self.metadata.get("name"),
                    "labels": {**(self.metadata.get("labels", {}) or {}), **labels},
                    "annotations": {**self.metadata.get("annotations", {}), **annotations},
                }
            ),
            "spec": self.spec.to_k8s(context=context, image=image, env_vars=env_vars),
        }


class RayJobSpec(dg.PermissiveConfig):
    """[RayJob spec](https://ray-project.github.io/kuberay/reference/api/#rayjobspec) configuration options. A few sensible defaults are provided for convenience."""

    active_deadline_seconds: int = 60 * 60 * 24  # 24 hours
    backoff_limit: int = 0
    ray_cluster_spec: RayClusterSpec | None = Field(default_factory=RayClusterSpec)
    submitter_pod_template: dict[str, Any] | None = None
    submitter_config: dict[str, Any] | None = None
    metadata: dict[str, Any] | None = None
    cluster_selector: dict[str, str] | None = None
    managed_by: str | None = None
    deletion_strategy: dict[str, Any] | None = Field(
        default_factory=lambda: {"onFailure": {"policy": "DeleteCluster"}, "onSuccess": {"policy": "DeleteCluster"}}
    )
    runtime_env_yaml: str | None = None
    job_id: str | None = None
    submission_mode: Literal["K8sJobMode", "HTTPMode", "InteractiveMode", "SidecarMode"] = "K8sJobMode"
    entrypoint_resources: str | None = None
    entrypoint_num_cpus: float | None = None
    entrypoint_memory: float | None = None
    entrypoint_num_gpus: float | None = None
    ttl_seconds_after_finished: int | None = 5 * 60  # 5 minutes
    shutdown_after_job_finishes: bool = True
    suspend: bool | None = None

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""
        return remove_none_from_dict(
            {
                "activeDeadlineSeconds": self.active_deadline_seconds,
                "backoffLimit": self.backoff_limit,
                "submitterPodTemplate": self.submitter_pod_template,
                "submitterConfig": self.submitter_config,
                "metadata": self.metadata,
                "clusterSelector": self.cluster_selector,
                "managedBy": self.managed_by,
                "deletionStrategy": self.deletion_strategy,
                "runtimeEnvYAML": self.runtime_env_yaml,
                "jobId": self.job_id,
                "submissionMode": self.submission_mode,
                "entrypointResources": self.entrypoint_resources,
                "entrypointNumCpus": self.entrypoint_num_cpus,
                "entrypointMemory": self.entrypoint_memory,
                "entrypointNumGpus": self.entrypoint_num_gpus,
                "ttlSecondsAfterFinished": self.ttl_seconds_after_finished,
                "shutdownAfterJobFinishes": self.shutdown_after_job_finishes,
                "suspend": self.suspend,
                "rayClusterSpec": self.ray_cluster_spec.to_k8s(context=context, image=image, env_vars=env_vars)
                if self.ray_cluster_spec is not None
                else None,
            }
        )


class RayJobConfig(dg.Config):
    kind: str = "RayJob"
    api_version: str = "ray.io/v1"
    metadata: dict[str, Any] = Field(
        default_factory=dict,
        description="Kubernetes metadata, except the name field can be omitted. In this case it will be generated by `dagster-ray`.",
    )
    spec: RayJobSpec = Field(default_factory=RayJobSpec)

    @property
    def namespace(self) -> str:
        return self.metadata.get("namespace", "default")

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        labels: Mapping[str, str] | None = None,
        annotations: Mapping[str, str] | None = None,
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""

        labels = labels or {}
        annotations = annotations or {}

        return {
            "apiVersion": self.api_version,
            "kind": self.kind,
            "metadata": remove_none_from_dict(
                {
                    "name": self.metadata.get("name"),
                    "labels": {**(self.metadata.get("labels", {}) or {}), **labels},
                    "annotations": {**self.metadata.get("annotations", {}), **annotations},
                }
            ),
            "spec": self.spec.to_k8s(
                context=context,
                image=image,
                env_vars=env_vars,
            ),
        }


class MatchDagsterLabels(dg.Config):
    cluster_sharing: bool = Field(default=True, description="Whether to match on `dagster/cluster-sharing=true` label.")
    code_location: bool = Field(
        default=True,
        description="Whether to match on `dagster/code-location` label. The value will be taken from the current Dagster code location.",
    )
    resource_key: bool = Field(
        default=True,
        description="Whether to match on `dagster/resource-key` label. The value will be taken from the current Dagster resource key.",
    )
    git_sha: bool = Field(
        default=True,
        description="Whether to match on `dagster/git-sha` label. The value will be taken from `DAGSTER_CLOUD_GIT_SHA` environment variable.",
    )  # TODO: we really should have common env vars for this, not just Dagster Plus specific
    run_id: bool = Field(
        default=False,
        description="Whether to match on `dagster/run-id` label. The value will be taken from the current Dagster run ID.",
    )


DEFAULT_CLUSTER_SHARING_TTL_SECONDS = 30 * 60.0


class ClusterSharing(dg.Config):
    """Defines the strategy for sharing `RayCluster` resources with other Dagster steps.

    By default, the cluster is expected to be created by Dagster during one of the previously executed steps.

    !!!note

        Cluster sharing uses the Kubernetes Lease API (`coordination.k8s.io`) for leader election
        to coordinate cluster creation across parallel steps. The Dagster ServiceAccount must have
        `create`, `get`, and `delete` permissions on `leases` in the `coordination.k8s.io` API group.
    """

    enabled: bool = Field(default=False, description="Whether to enable sharing of RayClusters.")
    match_dagster_labels: MatchDagsterLabels = Field(
        default_factory=MatchDagsterLabels, description="Configuration for matching on Dagster-generated labels."
    )
    match_labels: dict[str, str] | None = Field(
        default=None, description="Additional user-provided labels to match on."
    )
    ttl_seconds: float = Field(
        default=DEFAULT_CLUSTER_SHARING_TTL_SECONDS,
        description="Time to live for the lock placed on the `RayCluster` resource, marking it as in use by the current Dagster step.",
    )



from __future__ import annotations

import asyncio
import logging
import random
import string
import threading
import time
from collections.abc import AsyncIterator, Generator, Iterator
from contextlib import contextmanager
from functools import cached_property, partial
from typing import TYPE_CHECKING, Any, TypedDict, cast

import dagster as dg
import dagster._check as check
from dagster import (
    PipesContextInjector,
    PipesEnvContextInjector,
    PipesMessageHandler,
    PipesMessageReader,
    PipesSession,
    open_pipes_session,
)
from dagster._core.definitions.resource_annotation import TreatAsResourceParam
from dagster._core.errors import DagsterExecutionInterruptedError
from dagster._core.pipes.client import PipesClientCompletedInvocation
from dagster._core.pipes.context import PipesLaunchedData
from dagster._core.pipes.utils import (
    _join_thread,
    extract_message_or_forward_to_stdout,
)
from dagster_pipes import PipesDefaultMessageWriter, PipesExtras, PipesParams
from typing_extensions import NotRequired

from dagster_ray._base.utils import get_dagster_tags
from dagster_ray.types import OpOrAssetExecutionContext

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ray.job_submission import JobStatus, JobSubmissionClient


PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY = "ray_address"
PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY = "ray_job_id"


def tail_logs(tailer: AsyncIterator[str]) -> Generator[str | None, None, None]:
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    while True:
        try:
            yield loop.run_until_complete(tailer.__anext__())  # type: ignore
        except (StopAsyncIteration, ValueError):
            break
        except Exception:
            logger.exception("Encountered exception during ray job log tailing")
            yield None


class PipesRayJobMessageReader(PipesMessageReader):
    f"""
    Dagster Pipes message reader for receiving messages from a Ray job.
    Will extract Dagster events and forward the rest to stdout.

    Waits for PipesSession.report_launched to be called with `{PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY}` and `{PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY}` keys
    populated in `extras` in order to start message reading.

    Args:
        job_submission_client_kwargs: additional arguments for [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient]. The `address` value is expected to be set via `PipesSession.report_launched`, while other arguments can be set via this parameter.
    """

    _client: JobSubmissionClient | None
    _job_id: str | None

    def __init__(self, job_submission_client_kwargs: dict[str, Any] | None = None):
        self._job_submission_client_kwargs = job_submission_client_kwargs
        self._thread: threading.Thread | None = None
        self.session_closed = threading.Event()
        self._job_id = None
        self._client = None
        self.thread_ready = threading.Event()

        self.completed = threading.Event()

    @property
    def client(self) -> JobSubmissionClient:
        if self._client is None:
            raise RuntimeError(
                "client is only available after it has been seeded in the Pipes client via PipesSession.report_launched"
            )
        else:
            return self._client

    @property
    def job_id(self) -> str:
        if self._job_id is None:
            raise RuntimeError(
                "job_id is only available after it has been seeded in the Pipes client via PipesSession.report_launched"
            )
        else:
            return self._job_id

    def on_launched(self, launched_payload: PipesLaunchedData) -> None:
        from ray.job_submission import JobSubmissionClient

        self._client = JobSubmissionClient(
            address=launched_payload["extras"][PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY],
            **(self._job_submission_client_kwargs or {}),
        )
        self._job_id = launched_payload["extras"][PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY]

        super().on_launched(launched_payload)

    def messages_are_readable(self, params: PipesParams) -> bool:
        return self._job_id is not None and self._client is not None

    @contextmanager
    def read_messages(self, handler: PipesMessageHandler) -> Iterator[PipesParams]:
        try:
            self._thread = threading.Thread(
                target=partial(self.handle_job_logs, handler),
                daemon=True,
            )
            self._thread.start()
            yield {PipesDefaultMessageWriter.STDIO_KEY: PipesDefaultMessageWriter.STDOUT}
        finally:
            self.terminate(handler)

    def terminate(self, handler: PipesMessageHandler) -> None:
        self.session_closed.set()

        if self._thread is not None:
            _join_thread(self._thread, "ray job logs tailing")
            if self._job_id is None:
                handler._context.log.warning(
                    "[Pipes] PipesRayJobMessageReader._job_id is not set. It's expected to be set via PipesSession.report_launched."
                )

    def no_messages_debug_text(self) -> str:
        return f"Tried to read messages from Ray Job logs, but no messages were received. Make sure PipesSession.report_launched has been called with `{PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY}` set under `extras` field."

    def handle_job_logs(
        self,
        handler: PipesMessageHandler,
    ):
        try:
            while (self._job_id is None or self._client is None) and not self.session_closed.is_set():
                time.sleep(0.1)

            self.thread_ready.set()

            if self.session_closed.is_set():
                handler._context.log.warning(
                    "[pipes] PipesSession.report_launched was not called before session was closed."
                )
                return
            else:
                handler._context.log.debug("[pipes] PipesRayJobMessageReader is ready to begin reading ray job logs")

            assert self._job_id is not None and self._client is not None, (
                "Job ID and client are required for tailing logs. Expected on_launched to set them by this line."
            )

            session_closed_at = None

            tailer = self.client.tail_job_logs(job_id=self.job_id)

            for chunk in tail_logs(tailer):
                if chunk:
                    for log_line in chunk.split("\n"):
                        if log_line:
                            extract_message_or_forward_to_stdout(handler, log_line)

                if self.session_closed.is_set():
                    if session_closed_at is None:
                        session_closed_at = time.time()

                    if time.time() - session_closed_at > 30:  # wait for 30 seconds to flush all logs
                        self.completed.set()
                        return
        finally:
            self.completed.set()


class SubmitJobParams(TypedDict):
    """Parameters passed to [`JobSubmissionClient.submit_job`][ray.job_submission.JobSubmissionClient.submit_job]."""

    entrypoint: str
    runtime_env: NotRequired[dict[str, Any]]
    metadata: NotRequired[dict[str, str]]
    submission_id: NotRequired[str]
    entrypoint_num_cpus: NotRequired[float]
    entrypoint_num_gpus: NotRequired[float]
    entrypoint_memory: NotRequired[int]
    entrypoint_resources: NotRequired[dict[str, float]]
    entrypoint_label_selector: NotRequired[dict[str, str]]


class EnrichedSubmitJobParams(TypedDict):
    """Parameters passed to [`JobSubmissionClient.submit_job`][ray.job_submission.JobSubmissionClient.submit_job].

    This dictionary is typically constructed from [`SubmitJobParams`][dagster_ray.core.pipes.SubmitJobParams]
        with additional parameters injected by `dagster-ray`.
    """

    entrypoint: str
    runtime_env: dict[str, Any]
    metadata: dict[str, str]
    submission_id: str
    entrypoint_num_cpus: NotRequired[float]
    entrypoint_num_gpus: NotRequired[float]
    entrypoint_memory: NotRequired[int]
    entrypoint_resources: NotRequired[dict[str, float]]
    entrypoint_label_selector: NotRequired[dict[str, str]]


def generate_job_id() -> str:
    letters = string.ascii_lowercase
    random_letters = "".join(random.choice(letters) for i in range(10))
    return f"pipes-{random_letters}"


def _enrich_submit_job_params(
    context: OpOrAssetExecutionContext,
    session: PipesSession,
    submit_job_params: SubmitJobParams,
) -> EnrichedSubmitJobParams:
    """Merge Pipes env vars, Dagster tags, and a submission ID into job params."""
    runtime_env = submit_job_params.get("runtime_env", {})
    metadata = submit_job_params.get("metadata", {})

    env_vars = session.get_bootstrap_env_vars()

    metadata.update(get_dagster_tags(context))
    runtime_env["env_vars"] = runtime_env.get("env_vars", {})
    runtime_env["env_vars"].update(env_vars)

    submit_job_params["metadata"] = metadata
    submit_job_params["runtime_env"] = runtime_env
    submit_job_params["submission_id"] = submit_job_params.get("submission_id", generate_job_id())

    return cast(EnrichedSubmitJobParams, submit_job_params)


def _wait_for_job_completion(
    client: JobSubmissionClient,
    context: OpOrAssetExecutionContext,
    job_id: str,
    poll_interval: float,
) -> JobStatus:
    """Poll a Ray job until it reaches a terminal status."""
    from ray.job_submission import JobStatus

    context.log.info(f"[pipes] Waiting for Ray job {job_id} to complete...")

    while True:
        status = client.get_job_status(job_id)

        if status.is_terminal():
            if status in {JobStatus.FAILED, JobStatus.STOPPED}:
                job_details = client.get_job_info(job_id)
                raise RuntimeError(
                    f"[pipes] Ray job {job_id} failed with status {status}. Message:\n{job_details.message}"
                )

            return status

        time.sleep(poll_interval)


class PipesRayJobClient(dg.PipesClient, TreatAsResourceParam):
    """A [Pipes](https://docs.dagster.io/guides/build/external-pipelines) client for running Ray jobs on remote clusters.

    Starts the job directly on the Ray cluster and reads the logs from the job.

    Uses [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient] to run and monitor the job. Learn more about it [here](https://docs.ray.io/en/latest/cluster/running-applications/job-submission/sdk.html#python-sdk-overview).

    Args:
        address: Ray dashboard HTTP address.
            If unspecified, connects to a local Ray cluster or uses the ``RAY_ADDRESS`` environment variable.
        create_cluster_if_needed: Whether to start a local Ray cluster automatically if one is not already running. Useful for local development and testing.
        headers: HTTP headers for Ray Dashboard requests. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        verify: Whether to verify TLS certificate. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        cookies: HTTP cookies for Ray Dashboard requests. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        metadata: Ray Job metadata. Passed to [`JobSubmissionClient`][ray.job_submission.JobSubmissionClient].
        context_injector: A context injector to use to inject
            context into the Ray job. Defaults to [`PipesEnvContextInjector`][dagster.PipesEnvContextInjector].
        message_reader: A message reader to use when reading Pipes messages.
            from the glue job run. Defaults to [`PipesRayJobMessageReader`][dagster_ray.core.pipes.PipesRayJobMessageReader].
        forward_termination: Whether to cancel the Ray job run when the Dagster process receives a termination signal.
        timeout: Timeout for various internal interactions with the Ray job.
        poll_interval: Interval at which to poll Ray for status updates.
            Is useful when running in a local environment.
        _force_create_local_cluster: Workaround for [ray-project/ray#62091](https://github.com/ray-project/ray/issues/62091): calls ``ray.init()`` before creating the ``JobSubmissionClient`` to ensure a local cluster exists.
    """

    def __init__(
        self,
        address: str | None = None,
        create_cluster_if_needed: bool = False,
        headers: dict[str, Any] | None = None,
        verify: str | bool = True,
        cookies: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
        context_injector: PipesContextInjector | None = None,
        message_reader: PipesMessageReader | None = None,
        forward_termination: bool = True,
        timeout: float = 600,
        poll_interval: float = 1,
        _force_create_local_cluster: bool = False,
    ):
        self._address = address
        self._create_cluster_if_needed = create_cluster_if_needed
        self.__force_create_local_cluster = _force_create_local_cluster
        self._headers = headers
        self._verify = verify
        self._cookies = cookies
        self._metadata = metadata
        self._context_injector = context_injector or PipesEnvContextInjector()
        self._message_reader = message_reader or PipesRayJobMessageReader(
            job_submission_client_kwargs={
                "headers": headers,
                "verify": verify,
                "cookies": cookies,
                "metadata": metadata,
            }
        )

        self.forward_termination = check.bool_param(forward_termination, "forward_termination")
        self.timeout = check.numeric_param(timeout, "timeout")
        self.poll_interval = check.numeric_param(poll_interval, "poll_interval")

    @cached_property
    def client(self) -> JobSubmissionClient:
        from ray.job_submission import JobSubmissionClient

        if self.__force_create_local_cluster:
            import ray

            if not ray.is_initialized():
                ray.init()

        return JobSubmissionClient(
            address=self._address,
            create_cluster_if_needed=self._create_cluster_if_needed,
            headers=self._headers,
            verify=self._verify,
            cookies=self._cookies,
            metadata=self._metadata,
        )

    def run(  # type: ignore
        self,
        *,
        context: OpOrAssetExecutionContext,
        submit_job_params: SubmitJobParams,
        extras: PipesExtras | None = None,
    ) -> PipesClientCompletedInvocation:
        """
        Execute a RayJob, enriched with the Pipes protocol.

        Args:
            context: Current Dagster op or asset context.
            submit_job_params: Parameters for [`JobSubmissionClient.submit_job`][ray.job_submission.JobSubmissionClient.submit_job].
            extras: Additional information to pass to the Pipes session, retrievable via [`PipesContext.get_extras`][dagster_pipes.PipesContext.get_extra].
        """

        with open_pipes_session(
            context=context,
            message_reader=self._message_reader,
            context_injector=self._context_injector,
            extras=extras,
        ) as session:
            enriched_submit_job_params = _enrich_submit_job_params(context, session, submit_job_params)

            job_id = self._start(context, session, enriched_submit_job_params)

            try:
                _wait_for_job_completion(self.client, context, job_id, self.poll_interval)
                return PipesClientCompletedInvocation(session, metadata={"Ray Job ID": job_id})

            except DagsterExecutionInterruptedError:
                if self.forward_termination:
                    context.log.warning(f"[pipes] Dagster process interrupted! Will terminate RayJob {job_id}.")
                    self._terminate(context, job_id)
                raise

    def get_dagster_tags(self, context: OpOrAssetExecutionContext) -> dict[str, str]:
        tags = get_dagster_tags(context)
        return tags

    def _start(
        self, context: OpOrAssetExecutionContext, session: PipesSession, submit_job_params: EnrichedSubmitJobParams
    ) -> str:
        submission_id = submit_job_params["submission_id"]

        context.log.info(f"[pipes] Starting Ray job {submission_id}...")

        job_id = self.client.submit_job(**submit_job_params)

        context.log.info(f"[pipes] Ray job {job_id} started.")

        session.report_launched(
            {
                "extras": {
                    PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY: job_id,
                    PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY: self.client.get_address(),
                }
            }
        )

        return job_id

    def _terminate(self, context: OpOrAssetExecutionContext, job_id: str) -> None:
        context.log.info(f"[pipes] Terminating RayJob {job_id} ...")
        self.client.stop_job(job_id)
        context.log.info(f"[pipes] Ray job {job_id} terminated.")



import dagster as dg

from dagster_ray.kuberay.ops import delete_kuberay_clusters_op


@dg.job(description="Deletes KubeRay `RayCluster` resources")
def delete_kuberay_clusters():
    delete_kuberay_clusters_op()



"""Leader election via Kubernetes Lease for coordinating shared cluster creation.

This module provides a Kubernetes Lease-based leader election mechanism to ensure
only one step creates a shared RayCluster when multiple parallel steps start simultaneously.
"""

from __future__ import annotations

import logging
import time
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from kubernetes.client import ApiClient

logger = logging.getLogger(__name__)


@dataclass
class LeaderElectionResult:
    """Result of a leader election attempt."""

    is_leader: bool
    lease_name: str
    holder_identity: str


class LeaderElection:
    """Kubernetes Lease-based leader election for coordinating shared cluster creation.

    Uses the `coordination.k8s.io/v1` `Lease` API to implement distributed leader election.
    """

    def __init__(
        self,
        holder_identity: str,
        api_client: ApiClient,
        lease_duration_seconds: int = 60,
        retry_period_seconds: float = 1.0,
        acquire_timeout_seconds: float = 60.0,
    ):
        """Initialize leader election.

        Args:
            holder_identity: Unique identifier for this process
            api_client: Kubernetes API client
            lease_duration_seconds: How long the lease is held before expiring
            retry_period_seconds: How often to retry acquiring the lease
            acquire_timeout_seconds: Total time to wait for leader election
        """
        from kubernetes import client

        self.api_client = api_client
        self._coordination_api = client.CoordinationV1Api(api_client=api_client)
        self._holder_identity = holder_identity
        self.lease_duration_seconds = lease_duration_seconds
        self.retry_period_seconds = retry_period_seconds
        self.acquire_timeout_seconds = acquire_timeout_seconds

    def acquire_or_wait(
        self,
        lease_name: str,
        namespace: str,
        resource_exists_check: Callable[[], bool],
    ) -> LeaderElectionResult:
        """Acquire leadership or wait for the resource to be created by the leader.

        Args:
            lease_name: Name of the lease to use for coordination
            namespace: Kubernetes namespace
            resource_exists_check: Callable that returns True if the shared resource exists

        Returns:
            LeaderElectionResult indicating whether this process should create the resource
        """
        start_time = time.time()

        while time.time() - start_time < self.acquire_timeout_seconds:
            # Check if resource already exists
            if resource_exists_check():
                logger.debug("Shared resource already exists, not acquiring leadership")
                # Try to get the current lease holder for logging purposes
                current_holder = self._get_lease_holder(lease_name, namespace)
                return LeaderElectionResult(
                    is_leader=False,
                    lease_name=lease_name,
                    holder_identity=current_holder or "unknown",
                )

            # Try to become the leader
            if self._try_create_lease(lease_name, namespace):
                logger.info(f"Acquired leadership by creating lease {namespace}/{lease_name}")
                return LeaderElectionResult(
                    is_leader=True,
                    lease_name=lease_name,
                    holder_identity=self._holder_identity,
                )

            # Someone else is the leader, wait for resource to appear
            current_holder = self._get_lease_holder(lease_name, namespace)
            logger.debug(f"Lease {namespace}/{lease_name} held by {current_holder}, waiting for resource")
            time.sleep(self.retry_period_seconds)

        # Timeout - assume leadership to avoid deadlock
        logger.warning(f"Leader election timed out after {self.acquire_timeout_seconds}s, assuming leadership")
        return LeaderElectionResult(
            is_leader=True,
            lease_name=lease_name,
            holder_identity=self._holder_identity,
        )

    def _try_create_lease(self, name: str, namespace: str) -> bool:
        """Attempt to create a new lease. Returns True if successful (we're the leader)."""
        from kubernetes.client import ApiException
        from kubernetes.client.models import V1Lease, V1LeaseSpec, V1ObjectMeta

        lease = V1Lease(
            metadata=V1ObjectMeta(name=name, namespace=namespace),
            spec=V1LeaseSpec(
                holder_identity=self._holder_identity,
                lease_duration_seconds=self.lease_duration_seconds,
                acquire_time=datetime.now(timezone.utc),
            ),
        )

        try:
            self._coordination_api.create_namespaced_lease(namespace, lease)
            return True
        except ApiException as e:
            if e.status == 409:  # Already exists - someone else is the leader
                return False
            raise

    def _get_lease_holder(self, name: str, namespace: str) -> str | None:
        """Get the current holder identity from an existing lease."""
        from typing import cast

        from kubernetes.client import ApiException
        from kubernetes.client.models import V1Lease

        try:
            lease = cast(V1Lease, self._coordination_api.read_namespaced_lease(name=name, namespace=namespace))
            return lease.spec.holder_identity if lease.spec else None
        except ApiException as e:
            if e.status == 404:
                return None
            raise

    def _delete_lease(self, name: str, namespace: str) -> None:
        """Delete a lease."""
        from kubernetes.client import ApiException

        try:
            self._coordination_api.delete_namespaced_lease(name=name, namespace=namespace)
        except ApiException as e:
            if e.status != 404:
                raise

    def release(self, lease_name: str, namespace: str) -> None:
        """Release leadership by deleting the lease."""
        self._delete_lease(lease_name, namespace)
        logger.debug(f"Released lease {namespace}/{lease_name}")



import os
from collections.abc import Generator
from datetime import datetime, timezone
from typing import cast

import dagster as dg

from dagster_ray._base.cluster_sharing_lock import ClusterSharingLock
from dagster_ray.configs import (
    DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_DEFAULT_VALUE,
    DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_ENV_VAR,
    DAGSTER_RAY_NAMESPACES_ENV_VAR,
)
from dagster_ray.kuberay.client.raycluster.client import RayClusterClient
from dagster_ray.kuberay.jobs import delete_kuberay_clusters
from dagster_ray.kuberay.ops import DeleteKubeRayClustersConfig, RayClusterRef


@dg.sensor(job=delete_kuberay_clusters, minimum_interval_seconds=5 * 60)
def cleanup_expired_kuberay_clusters(
    context: dg.SensorEvaluationContext,
    raycluster_client: dg.ResourceParam[RayClusterClient],
) -> Generator[dg.RunRequest | dg.SkipReason, None, None]:
    f"""A Dagster sensor that monitors shared `RayCluster` resources created by the current code location and submits jobs to delete clusters that either:
        - use [Cluster Sharing](../tutorial/#cluster-sharing) (`dagster/cluster-sharing=true`) and have expired
        - are older than `DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS` (defaults to 4 hours)

    By default it monitors the `ray` namespace. This can be configured by setting `{DAGSTER_RAY_NAMESPACES_ENV_VAR}` (accepts a comma-separated list of namespaces)."""
    assert context.code_location_origin is not None

    found_any = False
    namespaces = os.environ.get(DAGSTER_RAY_NAMESPACES_ENV_VAR, "ray").split(",")
    expiration_seconds = int(
        os.environ.get(
            DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_ENV_VAR, DAGSTER_RAY_CLUSTER_EXPIRATION_SECONDS_DEFAULT_VALUE
        )
    )

    for namespace in namespaces:
        cluster_names = []
        for cluster in raycluster_client.list(
            namespace=namespace,
            label_selector=f"dagster/code-location={context.code_location_origin.location_name}",
        ).get("items", []):
            if cluster["metadata"].get("labels", {}).get("dagster/cluster-sharing") == "true":
                locks = ClusterSharingLock.parse_all_locks(
                    cast(dict[str, str], cluster.get("metadata", {}).get("annotations", {}))
                )
                alive_locks = ClusterSharingLock.get_alive_locks(locks)
                if not alive_locks:
                    context.log.info(
                        f"Found expired RayCluster with cluster sharing enabled: {cluster['metadata']['namespace']}/{cluster['metadata']['name']}"
                    )
                    cluster_names.append(cluster["metadata"]["name"])
            else:
                # check if the cluster age since creation time exceeds expiration_seconds

                # early exit for clusters with owners
                if cluster.get("metadata", {}).get("ownerReferences"):
                    continue

                if cluster.get("metadata", {}).get("creationTimestamp"):
                    cluster_age = datetime.now(timezone.utc) - datetime.strptime(
                        cluster["metadata"]["creationTimestamp"], "%Y-%m-%dT%H:%M:%SZ"
                    ).replace(tzinfo=timezone.utc)
                    if cluster_age.total_seconds() >= expiration_seconds:
                        context.log.info(
                            f"Found expired RayCluster (time since creation exceeds {expiration_seconds} seconds): {cluster['metadata']['namespace']}/{cluster['metadata']['name']}"
                        )
                        cluster_names.append(cluster["metadata"]["name"])

        if len(cluster_names) > 0:
            found_any = True
            yield dg.RunRequest(
                run_config=dg.RunConfig(
                    ops={
                        "delete_kuberay_clusters_op": DeleteKubeRayClustersConfig(
                            namespace=namespace,
                            clusters=[RayClusterRef(name=name) for name in cluster_names],
                        )
                    }
                )
            )

    if not found_any:
        yield dg.SkipReason(f"No expired RayClusters found in namespaces: {namespaces}")



import logging
from collections.abc import Sequence
from datetime import datetime, timedelta

from pydantic import BaseModel, Field, ValidationError
from typing_extensions import Self

logger = logging.getLogger(__name__)


class ClusterSharingLock(BaseModel):
    run_id: str = Field(description="The ID of the Dagster run that placed the lock.")
    key: str = Field(description="The key of the Dagster step that placed the lock, any unique identifier.")
    created_at: datetime = Field(description="The time at which the lock was created.")
    ttl_seconds: float = Field(description="Time to live for the lock after which it's considered expired.")

    @property
    def identifier(self) -> str:
        return f"{self.run_id}-{self.key}"

    @property
    def tag(self) -> str:
        return f"dagster/lock-{self.identifier}"[:63]  # 63 is the limit for k8s

    @property
    def expired_at(self) -> datetime:
        return self.created_at + timedelta(seconds=self.ttl_seconds)

    @property
    def is_expired(self) -> bool:
        return self.created_at + timedelta(seconds=self.ttl_seconds) < datetime.now()

    @classmethod
    def parse_all_locks(cls, data: dict[str, str]) -> Sequence[Self]:
        locks: list[Self] = []
        if data:
            for key, value in data.items():
                if key.startswith("dagster/lock-"):
                    try:
                        lock = cls.model_validate_json(value)
                        locks.append(lock)
                    except ValidationError:
                        logger.exception(
                            f"Invalid cluster sharing lock: {key}={value}. Consider updating `dagster-ray`."
                        )
        return locks

    @classmethod
    def get_alive_locks(cls, locks: Sequence[Self]) -> Sequence[Self]:
        return [lock for lock in locks if not lock.is_expired]



from __future__ import annotations

import re
from abc import ABC, abstractmethod

from pydantic import Field

from dagster_ray._base.constants import DEFAULT_DEPLOYMENT_NAME
from dagster_ray._base.resources import RayResource
from dagster_ray.kuberay.utils import get_k8s_object_name, k8s_service_fqdn
from dagster_ray.types import AnyDagsterContext


class BaseKubeRayResource(RayResource, ABC):
    image: str | None = Field(
        default=None,
        description="Image to inject into the `RayCluster` spec. Defaults to `dagster/image` run tag. Images already provided in the `RayCluster` spec won't be overridden.",
    )
    deployment_name: str = Field(
        default=DEFAULT_DEPLOYMENT_NAME,
        description="Dagster deployment name. Is used as a prefix for the Kubernetes resource name. Dagster Cloud variables are used to determine the default value.",
    )
    failure_tolerance_timeout: float = Field(
        default=0.0,
        description="The period in seconds to wait for the cluster to transition out of `failed` state if it reaches it. This state can be transient under certain conditions. With the default value of 0, the first `failed` state appearance will raise an exception immediately.",
    )
    poll_interval: float = Field(default=1.0, description="Poll interval for various API requests")

    @property
    @abstractmethod
    def namespace(self) -> str:
        raise NotImplementedError

    def resolve_hostname(self, service_name: str, namespace: str) -> str:
        """Resolve the host address from a Kubernetes service name and namespace.

        By default, returns the in-cluster FQDN (e.g. `my-svc.namespace.svc.cluster.local`).
        Override this method to customize the host address, for example to use a custom domain.

        The returned hostname is used to construct both the Ray Client address (`ray://<host>:<redis_port>`)
        and the dashboard URL (`http://<host>:<dashboard_port>`).

        Args:
            service_name: The Kubernetes service name for the Ray head node.
            namespace: The Kubernetes namespace.

        Returns:
            The hostname to use for connecting to the Ray cluster.

        Example:
            ```python
            class MyKubeRayCluster(KubeRayCluster):
                def resolve_hostname(self, service_name: str, namespace: str) -> str:
                    return f"{service_name}.{namespace}.company.com"
            ```
        """
        return k8s_service_fqdn(service_name, namespace)

    def _get_step_name(self, context: AnyDagsterContext) -> str:
        assert isinstance(context.run_id, str)
        assert context.dagster_run is not None

        # try to make the name as short as possible
        cluster_name_prefix = f"dg-{self.deployment_name.replace('-', '')[:8]}-{context.run_id[:8]}"

        dagster_user_email = context.dagster_run.tags.get("user")
        if dagster_user_email is not None:
            cluster_name_prefix += f"-{dagster_user_email.replace('.', '').replace('-', '').split('@')[0][:6]}"

        name_key = get_k8s_object_name(
            context.run_id,
            self.resource_uid,
        )

        step_name = f"{cluster_name_prefix}-{name_key}".lower()
        step_name = re.sub(r"[^-0-9a-z]", "-", step_name)

        return step_name
